import argparse
from tqdm import tqdm
import numpy as np 
import torch

import shutil
import sys
import os
sys.path.append('../../')
from GND.utils.print import cprint, lcprint
def get_parse():
    parser = argparse.ArgumentParser(description='CoLA: Self-Supervised Contrastive Learning for Anomaly Detection')
    parser.add_argument('--model', type=str, default='AAGNN_A')
    parser.add_argument('--dataset', type=str, default='Cora')  # "Cora", "Pubmed", "Citeseer"
    parser.add_argument('--lr', type=float)
    parser.add_argument('--weight_decay', type=float, default=0.0)
    parser.add_argument('--seed', type=int, default=1)
    parser.add_argument('--embedding_dim', type=int, default=64)
    parser.add_argument('--num_epoch', type=int)
    parser.add_argument('--subgraph_size', type=int, default=4)
    parser.add_argument('--readout', type=str, default='avg')  #max min avg  weighted_sum
    parser.add_argument('--auc_test_rounds', type=int)
    parser.add_argument('--num_workers', type=int, default=8)
    parser.add_argument('--negsamp_ratio', type=int, default=1)
    parser.add_argument('--device', type=int, default=0)
    parser.add_argument('--logdir', type=str, default='tmp')  #max min avg  weighted_sum
    parser.add_argument('--global_adg', type=bool, default=True)  #max min avg  weighted_sum
    args = parser.parse_args()
    
    if os.path.exists(args.logdir):
        shutil.rmtree(args.logdir)

    if args.lr is None:
        if args.dataset in ['Cora','Citeseer','Pubmed','Flickr']:
            args.lr = 1e-3
        elif args.dataset == 'ACM':
            args.lr = 5e-4
        elif args.dataset == 'BlogCatalog':
            args.lr = 3e-3
        else:
            args.lr = 1e-3

    if args.num_epoch is None:
        if args.dataset in ['Cora','Citeseer','Pubmed']:
            args.num_epoch = 100
        elif args.dataset in ['BlogCatalog','Flickr','ACM']:
            args.num_epoch = 400
        else:
            args.num_epoch = 10
            
    if args.dataset != 'ogbn-arxiv':
        args.auc_test_rounds = 256
    else:
        args.auc_test_rounds = 20
            
    return args



import dgl
import torch
import numpy as np

from .utils import move_start_node_fisrt

# TODO CoLA Sample
class BaseSubGraphSampling:
    r"""An abstract class for writing transforms."""

    def __call__(self, g, start_node):
        raise NotImplementedError

    def __repr__(self):
        return self.__class__.__name__ + "()"

class UniformNeighborSampling(BaseSubGraphSampling):
    r"""
    Uniform sampling Neighbors to generate subgraph.
    """
    def __init__(self, length=4):
        self.length = 4

    def __call__(self, g, start_nodes):
        rwl = []
        for node in start_nodes:
            pace = [node]
            successors = g.successors(node).numpy().tolist()
            # remove node and shuffle
            successors.remove(node)
            np.random.shuffle(successors)
            pace += successors[:self.length-1]
            pace += [pace[0]] * max(0, self.length - len(pace))
            rwl.append(pace)
        return rwl

class CoLASubGraphSampling(BaseSubGraphSampling):
    r"""
    ''
    we adopt random walk with restart (RWR)
    as local subgraph sampling strategy due to its usability and efficiency.
    we fixed the size 𝑆 of the sampled subgraph (number of nodes in the subgraph) to 4.
    For isolated nodes or the nodes which belong to a community with a size smaller than
    the predetermined subgraph size, we sample the available nodes repeatedly until an
    overlapping subgraph with the set size is obtained."
    described in [CoLA Anomaly Detection on Attributed Networks via Contrastive Self-Supervised Learning](https://arxiv.org/abs/2103.00113)
    Parameters:
    -----------
    g: DGLGraph object

    start_nodes: a Tensor or array contain start node.
    Return:
    -------
    rwl: List[List]
    examples:
    ---------
        ```python
        cola_sampler = CoLASubGraphSampling()
        g = dgl.graph(([0, 1, 2, 3, 6], [1, 2, 3, 4, 0]))
        g = dgl.add_reverse_edges(g)
        g = dgl.add_self_loop(g)
        ans = cola_sampler(g, [1, 2, 3, 5])
        print(ans)
        # [[1, 0, 2, 3], [2, 1, 0, 6], [3, 1, 2, 0], [5, 5, 5, 5]]
        ```
    """

    def __init__(self, length=4):
        self.length = 4

    def __call__(self, g, start_nodes):
        """
        add self_loop to handle isolated nodes as soon as
        the nodes which belong to a community with a size smaller than
        it is a little different from author's paper.
        """
        # newg = dgl.remove_self_loop(g)
        # newg = dgl.add_self_loop(newg)
        # length is Very influential to the effect of the model, maybe caused "remote" neighbor is 
        # not "friendly" to Rebuild Properties.
        paces = dgl.sampling.random_walk(g, start_nodes, length=self.length*3, restart_prob=0)[0]
        rwl = []
        for start, pace in zip(start_nodes, paces):
            pace = pace.unique().numpy()
            np.random.shuffle(pace)
            pace = pace[: self.length].tolist()
            pace = move_start_node_fisrt(pace, start)
            pace += [pace[0]] * max(0, self.length - len(pace))
            rwl.append(pace)
        return rwl


if __name__ == "__main__":
    cola_sampler = CoLASubGraphSampling()
    g = dgl.graph(([0, 1, 2, 3, 6], [1, 2, 3, 4, 0]))
    g = dgl.add_reverse_edges(g)
    # g = dgl.add_self_loop(g) for isolated nodes 
    ans = cola_sampler(g, [1, 2, 3, 5])
    print(ans)
    # [[1, 0, 2, 3], [2, 1, 0, 6], [3, 1, 2, 0], [5, 5, 5, 5]]
    subg = dgl.node_subgraph(g, ans[0])
    print(subg)


import torch
from ogb.nodeproppred import DglNodePropPredDataset
from scipy.stats import rankdata

def ranknorm(input_arr):
    r"""
    input_arr: np.ndarray like object.
    """
    return rankdata(input_arr, method='min') / len(input_arr)

def allclose(a, b, rtol=1e-4, atol=1e-4):
    return torch.allclose(a.float().cpu(),
            b.float().cpu(), rtol=rtol, atol=atol)

def move_start_node_fisrt(pace, start_node):
    """
    return a new pace in which the start node is in the first place.
    """
    if pace[0] == start_node:return pace
    for i in range(1, len(pace)):
        if pace[i] == start_node:
            pace[i] = pace[0]
            break
    pace[0] = start_node
    return pace

def is_bidirected(g):
    """Return whether the graph is a bidirected graph.
    A graph is bidirected if for any edge :math:`(u, v)` in :math:`G` with weight :math:`w`,
    there exists an edge :math:`(v, u)` in :math:`G` with the same weight.
    """
    src, dst = g.edges()
    num_nodes = g.num_nodes()

    # Sort first by src then dst
    idx_src_dst = src * num_nodes + dst
    perm_src_dst = torch.argsort(idx_src_dst, dim=0, descending=False)
    src1, dst1 = src[perm_src_dst], dst[perm_src_dst]

    # Sort first by dst then src
    idx_dst_src = dst * num_nodes + src
    perm_dst_src = torch.argsort(idx_dst_src, dim=0, descending=False)
    src2, dst2 = src[perm_dst_src], dst[perm_dst_src]

    return allclose(src1, dst2) and allclose(src2, dst1)

def load_ogbn_arxiv():
    data = DglNodePropPredDataset(name="ogbn-arxiv")
    graph, _ = data[0]
    # add reverse edges
    srcs, dsts = graph.all_edges()
    graph.add_edges(dsts, srcs)
    # add self-loop
    print(f"Total edges before adding self-loop {graph.number_of_edges()}")
    graph = graph.remove_self_loop().add_self_loop()
    print(f"Total edges after adding self-loop {graph.number_of_edges()}")
    assert is_bidirected(graph) == True
    return [graph]


    

r"""
cd CoLA
python main.py --dataset ACM
"""
if __name__ == "__main__":
    load_ogbn_arxiv()



import dgl
import torch
import numpy as np
import pandas as pd

from dgl.data import DGLDataset
from sklearn import preprocessing
from sklearn.metrics import roc_auc_score
from scipy.spatial.distance import euclidean
import scipy.sparse as sp
import  os
from dgl.data.utils import download

import scipy.io as sio
from dgl import backend as F
import sys
sys.path.append('..')
from GND.common.utils import is_bidirected, load_ogbn_arxiv#, load_ACM, load_BlogCatalog, load_Flickr
data_path = '../../data/'
#'BlogCatalog'  'Flickr' 'cora'  'citeseer' 'pubmed' 'ACM' 'ogbn-arxiv'
# TODO: add all datasets above.
def load_BlogCatalog():
    dataset=BlogCatalogGraphDataset(raw_dir=data_path)
    graph = dataset[0]
    # # add reverse edges
    # srcs, dsts = graph.all_edges()
    # graph.add_edges(dsts, srcs)
    # add self-loop
    print(f"Total edges before adding self-loop {graph.number_of_edges()}")
    graph = graph.remove_self_loop().add_self_loop()
    
    print(f"Total edges after adding self-loop {graph.number_of_edges()}")
    assert is_bidirected(graph) == True
    return [graph]

def load_Flickr():
    dataset=FlickerGraphDataset(raw_dir=data_path)
    graph = dataset[0]
    # # add reverse edges
    # srcs, dsts = graph.all_edges()
    # graph.add_edges(dsts, srcs)
    # add self-loop
    print(f"Total edges before adding self-loop {graph.number_of_edges()}")
    graph = graph.remove_self_loop().add_self_loop()
    print(f"Total edges after adding self-loop {graph.number_of_edges()}")
    assert is_bidirected(graph) == True
    return [graph]
    

def load_ACM():
    raise NotImplementedError

def split_auc(groundtruth, prob):
    r"""
    print the scoring(AUC) of the two types of anomalies separately.
    Parameter:
    ----------
    groundtruth: np.ndarray, Indicates whether this node is an injected anomaly node.
            0: normal node
            1: structural anomaly
            2: contextual anomaly

    prob: np.ndarray-like array saving the predicted score for every node
    Return:
    -------
    None
    """
    str_pos_idx = groundtruth == 1
    attr_pos_idx = groundtruth == 2
    norm_idx = groundtruth == 0

    str_data_idx = str_pos_idx | norm_idx
    attr_data_idx = attr_pos_idx | norm_idx

    str_data_groundtruth = groundtruth[str_data_idx]
    str_data_predict = prob[str_data_idx]

    attr_data_groundtruth = np.where(groundtruth[attr_data_idx] != 0, 1, 0)
    attr_data_predict = prob[attr_data_idx]

    s_score = roc_auc_score(str_data_groundtruth, str_data_predict)
    a_score = roc_auc_score(attr_data_groundtruth, attr_data_predict)
    final_score = roc_auc_score(np.where(groundtruth == 0, 0, 1), prob)

    print("structural anomaly score:", s_score)
    print("attribute anomaly score:", a_score)
    print("final anomaly score:", final_score)
    return final_score, a_score, s_score


class GraphNodeAnomalyDectionDataset(DGLDataset):
    r"""
    follow [CoLA](https://arxiv.org/abs/2103.00113)，inject Anomaly to the oral graph
    We fix p = 15
    and set q to 10, 15, 20, 5, 5, 20, 200 for
    BlogCatalog, Flickr, ACM, Cora, Citeseer, Pubmed and ogbn-arxiv, respectively.
    """

    def __init__(self, name="Cora", p=15, k=50, cola_preprocess_features=True, g_data=None, y_data=None):
        r"""
        Parameter
        ---------
        name:
        when name == 'custom', using custom data and please Specify custom data by g_data.
        and Specify label by y_data. [BlogCatalog, Flickr, Cora, Citeseer, Pubmed and ogbn-arxiv] is supported default follow CoLA.
        
        p and k :
        and anomaly injection hyperparameter follow CoLA

        cola_preprocess_features: 
        follow the same preprocess as CoLA, default:True

        g_data:
        Specify custom data by g_data.

        y_data:
        Specify custom label by g_data.

        """
        super().__init__(name=name)
        self.dataset_name = name
        self.cola_preprocess_features = cola_preprocess_features
        self.p = p
        self.q_map = {
            "BlogCatalog": 10,
            "Flickr": 15,
            "ACM": 20,
            "Cora": 5,
            "Citeseer": 5,
            "Pubmed": 20,
            "ogbn-arxiv": 200,
        }
        self.dataset_map = {
            "Cora": "dgl.data.CoraGraphDataset()",
            "Citeseer": "dgl.data.CiteseerGraphDataset()",
            "Pubmed": "dgl.data.PubmedGraphDataset()",
            "ogbn-arxiv":"load_ogbn_arxiv()",
            "ACM":"load_ACM()",
            "BlogCatalog":"load_BlogCatalog()",
            "Flickr":"load_Flickr()"
        }
        if self.dataset_name != 'custom':
            assert self.dataset_name in self.q_map and self.dataset_name in self.dataset_map, self.dataset_name
            self.q = self.q_map[name]
            self.k = k
            self.seed = 42
            self.dataset = eval(self.dataset_map[self.dataset_name])[0]
        else:
            self.dataset = g_data

        assert is_bidirected(self.dataset) == True
        self.init_anomaly_label(label=y_data)

        if self.dataset_name != 'custom' and y_data == None:
            print('inject_contextual_anomalies and inject_structural_anomalies')
            self.inject_contextual_anomalies()
            self.inject_structural_anomalies()

        if self.cola_preprocess_features:
            print('preprocess_features as CoLA')
            self.dataset.ndata['feat'] = self.preprocess_features(self.dataset.ndata['feat'])
    @property
    def num_anomaly(self):
        r"""
        anomaly_label: Indicates whether this node is an injected anomaly node.
            0: normal node
            1: structural anomaly
            2: contextual anomaly
        """
        return sum(self.dataset.ndata["anomaly_label"] != 0)

    @property
    def num_nodes(self):
        return self.dataset.num_nodes()

    @property
    def anomaly_label(self):
        return self.dataset.ndata["anomaly_label"]

    @property
    def anomalies_idx(self):
        anomalies = torch.where(self.anomaly_label != 0)[0].numpy()
        return anomalies

    @property
    def structural_anomalies_idx(self):
        anomalies = torch.where(self.anomaly_label == 1)[0].numpy()
        return anomalies

    @property
    def contextual_anomalies_idx(self):
        anomalies = torch.where(self.anomaly_label == 2)[0].numpy()
        return anomalies

    @property
    def normal_idx(self):
        nodes = torch.where(self.anomaly_label == 0)[0].numpy()
        return nodes

    @property
    def node_attr(self):
        return self.dataset.ndata["feat"]

    def set_node_attr(self, attr):
        self.dataset.ndata["feat"] = attr

    def set_anomaly_label(self, label):
        self.dataset.ndata["anomaly_label"] = label

    def init_anomaly_label(self, label=None):
        number_node = self.dataset.num_nodes()
        if label != None:
            self.dataset.ndata["anomaly_label"] = label# torch.zeros(number_node)
        else:
            self.dataset.ndata["anomaly_label"] = torch.zeros(number_node)

    def reset_anomaly_label(self):
        self.init_anomaly_label()

    def process(self):
        pass

    def inject_structural_anomalies(self):
        np.random.seed(self.seed)
        src, dst = self.dataset.edges()
        labels = self.anomaly_label
        p, q = self.p, self.q
        number_nodes = self.dataset.num_nodes()
        anomalies = set(torch.where(labels != 0)[0].numpy())

        new_src, new_dst = [], []
        # q cliques
        for i in range(q):
            q_list = []
            # selet p nodes
            for j in range(p):
                a = np.random.randint(number_nodes)
                while a in anomalies:
                    a = np.random.randint(number_nodes)
                q_list.append(a)
                anomalies.add(a)
                labels[a] = 1
            # make full connected
            for n1 in range(p):
                for n2 in range(n1 + 1, p):
                    new_src.extend([q_list[n1], q_list[n2]])
                    new_dst.extend([q_list[n2], q_list[n1]])

        src, dst = list(src.numpy()), list(dst.numpy())
        src.extend(new_src)
        dst.extend(new_dst)
        # update edges
        self.dataset.remove_edges(torch.arange(self.dataset.num_edges()))
        self.dataset.add_edges(src, dst)
        # print(self.dataset.num_edges())
        # BUG
        r"""
        dgl.DGLGraph.to_simple is not supported inplace
        """
        # self.dataset.to_simple()
        self.dataset = dgl.to_simple(self.dataset)
        # print(self.dataset.num_edges())
        self.set_anomaly_label(labels)
        print(
            "inject structural_anomalies numbers:", len(self.structural_anomalies_idx)
        )
        print("anomalies numbers:", len(self.anomalies_idx))

    def inject_contextual_anomalies(self):
        np.random.seed(self.seed)
        k = self.k
        attribute_anomalies_number = self.p * self.q
        normal_nodes_idx = self.normal_idx
        attribute_anomalies_idx = np.random.choice(
            normal_nodes_idx, size=attribute_anomalies_number, replace=False
        )
        all_attr = self.node_attr
        all_nodes_idx = list(range(self.dataset.num_nodes()))
        for aa_i in attribute_anomalies_idx:
            # random sample k nodes
            random_k_idx = np.random.choice(all_nodes_idx, size=k, replace=False)
            # cal the euclidean distance and replace the node attribute with \
            biggest_distance = 0
            biggest_attr = 0
            for i in random_k_idx:
                dis = euclidean(all_attr[aa_i], all_attr[i])
                if dis > biggest_distance:
                    biggest_distance, biggest_attr = dis, all_attr[i]
            # the node which has biggest one euclidean distance
            all_attr[aa_i] = biggest_attr

        self.set_node_attr(all_attr)
        labels = self.anomaly_label
        labels[attribute_anomalies_idx] = 2
        self.set_anomaly_label(labels)
        print(
            "inject contextual_anomalies numbers:", len(self.contextual_anomalies_idx)
        )
        print("anomalies numbers:", len(self.anomalies_idx))

    def preprocess_features(self, features):
        r"""Row-normalize feature matrix and convert to tuple representation
        copy from [CoLA]()
        """
        rowsum = np.array(features.sum(1))
        r_inv = np.power(rowsum, -1).flatten()
        r_inv[np.isinf(r_inv)] = 0.
        r_mat_inv = sp.diags(r_inv)
        features = r_mat_inv.dot(features)
        return torch.Tensor(features).float()

    def evalution(self, prediction):
        r"""
        print the scoring(AUC) of the two types of anomalies separately.
        Parameter:
        ----------
        prediction: np.ndarray-like array save the predicted score for every node
        Return:
        -------
        None
        """
        return split_auc(self.anomaly_label, prediction)

    def evaluation_multiround(self, predict_score_arr):
        '''
        Description
        -----------
        mulit-round result(as CoLA) evaluation.

        Parameter
        ---------
        predict_score_arr: node * num_round
        '''
        mean_predict_result = predict_score_arr.mean(1)
        std_predict_result = predict_score_arr.std(1)
        max_predict_result = predict_score_arr.max(1)
        min_predict_result = predict_score_arr.min(1)
        median_predict_result = np.median(predict_score_arr, 1)

        descriptions = {
            "mean": mean_predict_result,
            "std": std_predict_result,
            "-std": - std_predict_result,
            "mean+std": mean_predict_result + std_predict_result,
            "mean-std": mean_predict_result - std_predict_result,
            "mean-2std": mean_predict_result - 2*std_predict_result,
            "mean-3std": mean_predict_result - 3*std_predict_result,
            "mean+median": mean_predict_result + median_predict_result,
            "max": max_predict_result,
            "min": min_predict_result,
            "min-std": min_predict_result-std_predict_result,
            "median": median_predict_result,
        }
        for stat in descriptions:
            print("=" * 10 + stat + "=" * 10)
            self.evalution(descriptions[stat])

    def __getitem__(self, idx):
        return self.dataset

    def __len__(self):
        return 0

class ACMGraphDataset(DGLDataset):
    """ACM citation network dataset.It is a citation network where each paper 
    is regarded as a node on the network, and the links are the citation 
    relations among different papers. The attributes of each paper are 
    generated from the paper abstract.
    
    From:

    Parameters
    ----------
    raw_dir : str
        指定下载数据的存储目录或已下载数据的存储目录。默认: ~/.dgl/
    verbose : bool
        是否打印进度信息。
    """
    def __init__(self,
                 raw_dir=None,
                 verbose=True):
        super(ACMGraphDataset, self).__init__(name='ACM',
                                        raw_dir=raw_dir,
                                        verbose=verbose)
        self.raw_dir=raw_dir
        self.verbose=verbose

    def process(self):
        mat_path =self.raw_dir
        data_mat = sio.loadmat(mat_path)
        adj = data_mat['Network']
        feat = data_mat['Attributes']
        truth = data_mat['Label']
        truth = truth.flatten()
        self._g=dgl.from_scipy(adj)
        self._g.ndata['feat']=torch.from_numpy(feat.toarray()).to(torch.float32)
        self._g.ndata['label']=torch.from_numpy(truth).to(torch.float32)
        self.num_classes=len(np.unique(truth))

        if self.verbose:
            print('  NumNodes: {}'.format(self._g.number_of_nodes()))
            print('  NumEdges: {}'.format(self._g.number_of_edges()))
            print('  NumFeats: {}'.format(self._g.ndata['feat'].shape[1]))
            print('  NumClasses: {}'.format(self.num_classes))


    def __getitem__(self, idx):
        assert idx == 0, "这个数据集里只有一个图"
        return self.graph

    def __len__(self):
        return 1

class FlickerGraphDataset(DGLDataset):
    """Flickr is an image hosting and sharing website. Similar to BlogCatalog,
     users can follow each other and form a social network. Node attributes of 
     users are defined by their specified tags that reflect their interests.
    
    From:https://github.com/GRAND-Lab/CoLA/blob/main/raw_dataset/Flickr/Flickr.mat

    Parameters
    ----------
    raw_dir : str
        指定下载数据的存储目录或已下载数据的存储目录。默认: ~/.dgl/
    verbose : bool
        是否打印进度信息。
    """
    def __init__(self,
                 raw_dir=None,
                 verbose=True):
        super(FlickerGraphDataset, self).__init__(name='Flickr',
                                        raw_dir=raw_dir,
                                        verbose=verbose)

    def process(self):
        mat_path =self.raw_path + '.mat'
        data_mat = sio.loadmat(mat_path)
        adj = data_mat['Network']
        feat = data_mat['Attributes']
        feat = preprocessing.normalize(feat, axis=0)
        truth = data_mat['Label']
        truth = truth.flatten()
        self._g=dgl.from_scipy(adj)
        self._g.ndata['feat']=torch.from_numpy(feat.toarray()).to(torch.float32)
        self._g.ndata['label']=torch.from_numpy(truth).to(torch.float32)
        self.num_classes=len(np.unique(truth))

        if self.verbose:
            print('  NumNodes: {}'.format(self._g.number_of_nodes()))
            print('  NumEdges: {}'.format(self._g.number_of_edges()))
            print('  NumFeats: {}'.format(self._g.ndata['feat'].shape[1]))
            print('  NumClasses: {}'.format(self.num_classes))

    def __getitem__(self, idx):
        assert idx == 0, "这个数据集里只有一个图"
        return self._g

    def __len__(self):
        return 1

class BlogCatalogGraphDataset(DGLDataset):
    """BlogCatalog is a blog sharing web- site. The bloggers in blogcatalog 
    can follow each other forming a social network. Users are associ- ated 
    with a list of tags to describe themselves and their blogs, which are regarded as node attributes.
    
    From:https://github.com/GRAND-Lab/CoLA/blob/main/raw_dataset/BlogCatalog/BlogCatalog.mat

    Parameters
    ----------
    raw_dir : str
        指定下载数据的存储目录或已下载数据的存储目录。默认: ~/.dgl/
    verbose : bool
        是否打印进度信息。
    """
    def __init__(self,
                 raw_dir=None,
                 verbose=True):
        super(BlogCatalogGraphDataset, self).__init__(name='BlogCatalog',
                                        raw_dir=raw_dir,
                                        verbose=verbose)

    def process(self):
        mat_path =self.raw_path + '.mat'
        
        data_mat = sio.loadmat(mat_path)
        adj = data_mat['Network']
        feat = data_mat['Attributes']
        feat = preprocessing.normalize(feat, axis=0)
        truth = data_mat['Label']
        truth = truth.flatten()
        self._g=dgl.from_scipy(adj)
        self._g.ndata['feat']=torch.from_numpy(feat.toarray()).to(torch.float32)
        self._g.ndata['label']=torch.from_numpy(truth).to(torch.float32)
        self.num_classes=len(np.unique(truth))

        if self.verbose:
            print('  NumNodes: {}'.format(self._g.number_of_nodes()))
            print('  NumEdges: {}'.format(self._g.number_of_edges()))
            print('  NumFeats: {}'.format(self._g.ndata['feat'].shape[1]))
            print('  NumClasses: {}'.format(self.num_classes))

    def __getitem__(self, idx):
        assert idx == 0, "这个数据集里只有一个图"
        return self._g

    def __len__(self):
        return 1


def test_cutom_dataset():
    my_g = dgl.data.CoraGraphDataset()[0]
    label = torch.ones(my_g.num_nodes())
    dataset = GraphNodeAnomalyDectionDataset('custom', g_data=my_g, y_data=label)
    print("num_anomaly:", dataset.num_anomaly)
    print("anomaly_label", dataset.anomaly_label)

if __name__ == "__main__":
    test_cutom_dataset()
    data_path = '../data/'
    well_test_dataset = ["Cora", "Pubmed", "Citeseer","BlogCatalog","Flickr", "ogbn-arxiv"]
    num_nodes_list = []
    num_edges_list = []
    num_anomaly_list = []
    num_attributes_list = []
    random_evaluation_list = []
    
    for data_name in well_test_dataset:
        print("\ndataset:", data_name)
        dataset = GraphNodeAnomalyDectionDataset(data_name)
        print("num_anomaly:", dataset.num_anomaly)
        print("anomaly_label", dataset.anomaly_label)
        rand_ans = np.random.rand(dataset.num_nodes)
        _, _, final_score = dataset.evalution(rand_ans)
        num_nodes_list.append(dataset.num_nodes)
        num_edges_list.append(dataset.dataset.num_edges())
        num_anomaly_list.append(dataset.num_anomaly.item())
        num_attributes_list.append(dataset.dataset.ndata['feat'].shape[1])
        random_evaluation_list.append(final_score)


    dataset_info = pd.DataFrame({
        "well_test_dataset":well_test_dataset,
        "num_nodes":num_nodes_list,
        "num_edges":num_edges_list,
        "num_anomaly": num_anomaly_list,
        "num_attributes": num_attributes_list,
        "random_evaluation":random_evaluation_list
    })
    print(dataset_info)


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from sklearn.metrics import roc_auc_score
from scipy.spatial.distance import euclidean
import scipy.sparse as spp
from tqdm import tqdm
from torch.utils.tensorboard import SummaryWriter

class AAGNN_M(nn.Module):
    def __init__(self,):
        super().__init__()
    def fit(self, dataset, args):
        graph = dataset[0]
        features = graph.ndata['feat']
        print(graph)
        print('features shape:', features.shape)
        if torch.cuda.is_available():
            device = torch.device("cuda:" + str(args.device))
        else:
            device = torch.device("cpu")
        features = features.to(device)
        model = AAGNN_M_base(graph, features.shape[1], 256, device)
        model = model.to(device)
        opt = torch.optim.Adam(model.parameters(), lr=args.lr)
        #获取伪标签下的正常样本
        mask = model.mask_label(features, 0.5)
        writer = SummaryWriter(log_dir=args.logdir)
        model.train()
        for epoch in range(args.num_epoch):
            out = model(features)
            loss = model.loss_fun(out, mask, model, 0.0001, device)
            opt.zero_grad()
            loss.backward()
            opt.step()
            predict_score = model.anomaly_score(out)
            print("Epoch:", '%04d' % (epoch), "train_loss=", "{:.5f}".format(loss.item(
            )))
            writer.add_scalars(
                "loss",
                {"loss": loss},
                epoch,
            )
            final_score, a_score, s_score = dataset.evalution(predict_score)
            writer.add_scalars(
                "auc",
                {"final": final_score, "structural": s_score, "attribute": a_score},
                epoch,
            )
            writer.flush()

class AAGNN_M_base(nn.Module):
    def __init__(self, g, in_feats, out_feats, device):
        super().__init__()
        self.line = nn.Linear(in_feats, out_feats).to(device)
        #生成邻接矩阵，A的尺寸为(n_node, n_node)
        self.A = torch.zeros((len(g.nodes().numpy()), len(g.nodes().numpy()))).to(device)
        us = g.edges()[0].numpy()
        vs = g.edges()[1].numpy()
        for u, v in zip(us, vs):
            self.A[u][v] = self.A[v][u] = 1.0
            self.A[u][u] = self.A[v][v] = 1.0
        #得到每个节点的度数矩阵，B的尺寸为(n_node, hid_feats)
        self.B = torch.sum(self.A, dim=1).reshape(-1, 1)

    def forward(self, inputs):
        #进行线性映射到低维空间，z的尺寸为(n_node, hid_feats)
        z = self.line(inputs)
        #A * z的结果矩阵尺寸为(n_node, hid_feats)，和度矩阵B对应点除，得到平均值
        h = z - (torch.mm(self.A, z)/ self.B)
        #最后非线性激活函数映射输出
        return F.relu(h)
    #计算伪标签，伪标签为正样本的mask下
    def mask_label(self, inputs, p):
        with torch.no_grad():
            z = self.line(inputs)
            #得到所有节点特征的均值矩阵
            c = torch.mean(z, dim=0)
            #计算距离
            dis = torch.sum((z - c) * (z - c), dim=1)
            best_min_dis = list(dis.cpu().data.numpy())
            #从小到大排序
            best_min_dis.sort()
            #得到距离阈值
            threshold = best_min_dis[int(len(best_min_dis) * p)]
            mask = (dis <= threshold)
            #返回mask，为true表示伪标签是正样本
            return mask
            
    #损失函数
    def loss_fun(self, out, mask, model, super_param, device):
        #得到所有节点特征的均值矩阵
        c = torch.mm(torch.ones(out.shape[0], 1).to(device), torch.mean(out, dim=0).reshape(1, -1))
        #计算所有节点的误差
        loss_matrix = torch.sum((out - c) * (out - c), dim=1)[mask]
        #取均值误差
        loss = torch.mean(loss_matrix, dim=0)

        l2_reg = torch.tensor(0.).to(device)#L2正则项
        for param in model.parameters():
            l2_reg += torch.norm(param)
        return loss + (super_param * l2_reg/2)

    #计算异常分数
    def anomaly_score(self, out):
        s = torch.sum(out * out, dim=1)
        return s.cpu().data.numpy()





from setuptools import setup, find_packages
setup(name='GND',
      version='0.0.5',
      description='GraphAnomalyDection Benchmarking using DGL',
      author='fmc123653',
      author_email='fmc0315@126.com', #这里务必填写正确的email格式
      packages=find_packages(),  # 表示你要封装的包，find_packages用于系统自动从当前目录开始找包
      license="apache 3.0"
      )